@@ -3,31 +3,26 @@
 
 #include "set.h"
 
-#include <sys/stat.h>
-#include <unistd.h>
-
 #include <algorithm>
 #include <cerrno>
-#include <cstddef>
-#include <cstdlib>
-#include <cstring>
 #include <cwchar>
-#include <iterator>
+#include <functional>
+#include <map>
 #include <memory>
-#include <set>
 #include <string>
+#include <utility>
 #include <vector>
 
 #include "../builtin.h"
 #include "../common.h"
 #include "../env.h"
+#include "../event.h"
 #include "../expand.h"
 #include "../fallback.h"  // IWYU pragma: keep
 #include "../history.h"
 #include "../io.h"
+#include "../maybe.h"
 #include "../parser.h"
-#include "../proc.h"
-#include "../wcstringutil.h"
 #include "../wgetopt.h"
 #include "../wutil.h"  // IWYU pragma: keep
 
@@ -61,30 +56,29 @@ enum {
 // (REQUIRE_ORDER) option for flag parsing. This is not typical of most fish commands. It means
 // we stop scanning for flags when the first non-flag argument is seen.
 static const wchar_t *const short_options = L"+:LSUaefghlnpqux";
-static const struct woption long_options[] = {{L"export", no_argument, nullptr, 'x'},
-                                              {L"global", no_argument, nullptr, 'g'},
-                                              {L"function", no_argument, nullptr, 'f'},
-                                              {L"local", no_argument, nullptr, 'l'},
-                                              {L"erase", no_argument, nullptr, 'e'},
-                                              {L"names", no_argument, nullptr, 'n'},
-                                              {L"unexport", no_argument, nullptr, 'u'},
-                                              {L"universal", no_argument, nullptr, 'U'},
-                                              {L"long", no_argument, nullptr, 'L'},
-                                              {L"query", no_argument, nullptr, 'q'},
-                                              {L"show", no_argument, nullptr, 'S'},
-                                              {L"append", no_argument, nullptr, 'a'},
-                                              {L"prepend", no_argument, nullptr, 'p'},
-                                              {L"path", no_argument, nullptr, opt_path},
-                                              {L"unpath", no_argument, nullptr, opt_unpath},
-                                              {L"help", no_argument, nullptr, 'h'},
-                                              {nullptr, 0, nullptr, 0}};
+static const struct woption long_options[] = {{L"export", no_argument, 'x'},
+                                              {L"global", no_argument, 'g'},
+                                              {L"function", no_argument, 'f'},
+                                              {L"local", no_argument, 'l'},
+                                              {L"erase", no_argument, 'e'},
+                                              {L"names", no_argument, 'n'},
+                                              {L"unexport", no_argument, 'u'},
+                                              {L"universal", no_argument, 'U'},
+                                              {L"long", no_argument, 'L'},
+                                              {L"query", no_argument, 'q'},
+                                              {L"show", no_argument, 'S'},
+                                              {L"append", no_argument, 'a'},
+                                              {L"prepend", no_argument, 'p'},
+                                              {L"path", no_argument, opt_path},
+                                              {L"unpath", no_argument, opt_unpath},
+                                              {L"help", no_argument, 'h'},
+                                              {}};
 
 // Hint for invalid path operation with a colon.
-#define BUILTIN_SET_MISMATCHED_ARGS _(L"%ls: You provided %d indexes but %d values\n")
-#define BUILTIN_SET_ERASE_NO_VAR _(L"%ls: Erase needs a variable name\n")
-#define BUILTIN_SET_ARRAY_BOUNDS_ERR _(L"%ls: Array index out of bounds\n")
+#define BUILTIN_SET_MISMATCHED_ARGS _(L"%ls: given %d indexes but %d values\n")
+#define BUILTIN_SET_ARRAY_BOUNDS_ERR _(L"%ls: array index out of bounds\n")
 #define BUILTIN_SET_UVAR_ERR \
-    _(L"%ls: Universal variable '%ls' is shadowed by the global variable of the same name.\n")
+    _(L"%ls: successfully set universal '%ls'; but a global by that name shadows it\n")
 
 static int parse_cmd_opts(set_cmd_opts_t &opts, int *optind,  //!OCLINT(high ncss method)
                           int argc, const wchar_t **argv, parser_t &parser, io_streams_t &streams) {
@@ -167,6 +161,20 @@ static int parse_cmd_opts(set_cmd_opts_t &opts, int *optind,  //!OCLINT(high ncs
                 return STATUS_INVALID_ARGS;
             }
             case '?': {
+                // Specifically detect `set -o` because people might be bringing over bashisms.
+                if (wcsncmp(argv[w.woptind - 1], L"-o", 2) == 0) {
+                    streams.err.append(
+                        L"Fish does not have shell options. See `help fish-for-bash-users`.\n");
+                    if (w.woptind < argc) {
+                        if (wcscmp(argv[w.woptind], L"vi") == 0) {
+                            // Tell the vi users how to get what they need.
+                            streams.err.append(L"To enable vi-mode, run `fish_vi_key_bindings`.\n");
+                        } else if (wcscmp(argv[w.woptind], L"ed") == 0) {
+                            // This should be enough for make ed users feel at home
+                            streams.err.append(L"?\n?\n?\n");
+                        }
+                    }
+                }
                 builtin_unknown_option(parser, streams, cmd, argv[w.woptind - 1]);
                 return STATUS_INVALID_ARGS;
             }
@@ -180,9 +188,8 @@ static int parse_cmd_opts(set_cmd_opts_t &opts, int *optind,  //!OCLINT(high ncs
     return STATUS_CMD_OK;
 }
 
-static int validate_cmd_opts(const wchar_t *cmd,
-                             const set_cmd_opts_t &opts,  //!OCLINT(npath complexity)
-                             int argc, parser_t &parser, io_streams_t &streams) {
+static int validate_cmd_opts(const wchar_t *cmd, const set_cmd_opts_t &opts, int argc,
+                             const wchar_t *argv[], parser_t &parser, io_streams_t &streams) {
     // Can't query and erase or list.
     if (opts.query && (opts.erase || opts.list)) {
         streams.err.append_format(BUILTIN_ERR_COMBO, cmd);
@@ -197,11 +204,14 @@ static int validate_cmd_opts(const wchar_t *cmd,
         return STATUS_INVALID_ARGS;
     }
 
-    // Variables can only have one scope.
+    // Variables can only have one scope...
     if (opts.local + opts.function + opts.global + opts.universal > 1) {
-        streams.err.append_format(BUILTIN_ERR_GLOCAL, cmd);
-        builtin_print_error_trailer(parser, streams.err, cmd);
-        return STATUS_INVALID_ARGS;
+        // ..unless we are erasing a variable, in which case we can erase from several in one go.
+        if (!opts.erase) {
+            streams.err.append_format(BUILTIN_ERR_GLOCAL, cmd);
+            builtin_print_error_trailer(parser, streams.err, cmd);
+            return STATUS_INVALID_ARGS;
+        }
     }
 
     // Variables can only have one export status.
@@ -234,7 +244,7 @@ static int validate_cmd_opts(const wchar_t *cmd,
     }
 
     if (argc == 0 && opts.erase) {
-        streams.err.append_format(BUILTIN_SET_ERASE_NO_VAR, cmd);
+        streams.err.append_format(BUILTIN_ERR_MISSING, cmd, L"--erase");
         builtin_print_error_trailer(parser, streams.err, cmd);
         return STATUS_INVALID_ARGS;
     }
@@ -323,17 +333,17 @@ static maybe_t<split_var_t> split_var_and_indexes(const wchar_t *arg, env_mode_f
                                                   const environment_t &vars,
                                                   io_streams_t &streams) {
     split_var_t res{};
-    const wchar_t *open_bracket = std::wcschr(arg, L'[');
-    size_t varname_len = open_bracket ? open_bracket - arg : wcslen(arg);
-    res.varname.assign(arg, varname_len);
+    wcstring argstr{arg};
+    auto open_bracket = argstr.find(L'[');
+    res.varname.assign(arg, open_bracket == wcstring::npos ? argstr.size() : open_bracket);
     res.var = vars.get(res.varname, mode);
-    if (!open_bracket) {
+    if (open_bracket == wcstring::npos) {
         // Common case of no bracket.
         return res;
     }
 
     long varsize = res.varsize();
-    const wchar_t *p = open_bracket + 1;
+    const wchar_t *p = arg + open_bracket + 1;
     while (*p != L']') {
         const wchar_t *end;
         long l_ind;
@@ -384,7 +394,7 @@ static maybe_t<split_var_t> split_var_and_indexes(const wchar_t *arg, env_mode_f
     return res;
 }
 
-/// Given a list of values and 1-based indexes, return a new list, with those elements removed.
+/// Given a list of values and 1-based indexes, return a new list with those elements removed.
 /// Note this deliberately accepts both args by value, as it modifies them both.
 static wcstring_list_t erased_at_indexes(wcstring_list_t input, std::vector<long> indexes) {
     // Sort our indexes into *descending* order.
@@ -430,8 +440,8 @@ static int builtin_set_list(const wchar_t *cmd, set_cmd_opts_t &opts, int argc,
     sort(names.begin(), names.end());
 
     for (const auto &key : names) {
-        const wcstring e_key = escape_string(key, 0);
-        streams.out.append(e_key);
+        wcstring out;
+        out.append(key);
 
         if (!names_only) {
             wcstring val;
@@ -454,14 +464,15 @@ static int builtin_set_list(const wchar_t *cmd, set_cmd_opts_t &opts, int argc,
                     shorten = true;
                     val.resize(60);
                 }
-                streams.out.append(L" ");
-                streams.out.append(val);
+                out.push_back(L' ');
+                out.append(val);
 
-                if (shorten) streams.out.push_back(get_ellipsis_char());
+                if (shorten) out.push_back(get_ellipsis_char());
             }
         }
 
-        streams.out.append(L"\n");
+        out.push_back(L'\n');
+        streams.out.append(out);
     }
 
     return STATUS_CMD_OK;
@@ -528,13 +539,14 @@ static void show_scope(const wchar_t *var_name, int scope, io_streams_t &streams
     const wchar_t *exportv = var->exports() ? _(L"exported") : _(L"unexported");
     const wchar_t *pathvarv = var->is_pathvar() ? _(L" a path variable") : L"";
     wcstring_list_t vals = var->as_list();
-    streams.out.append_format(_(L"$%ls: set in %ls scope, %ls,%ls with %d elements\n"), var_name,
+    streams.out.append_format(_(L"$%ls: set in %ls scope, %ls,%ls with %d elements"), var_name,
                               scope_name, exportv, pathvarv, vals.size());
     // HACK: PWD can be set, depending on how you ask.
     // For our purposes it's read-only.
     if (env_var_t::flags_for(var_name) & env_var_t::flag_read_only) {
-        streams.out.append(_(L"Variable is read-only\n"));
-    }
+        streams.out.append(_(L" (read-only)\n"));
+    } else
+        streams.out.push_back(L'\n');
 
     for (size_t i = 0; i < vals.size(); i++) {
         if (vals.size() > 100) {
@@ -546,7 +558,7 @@ static void show_scope(const wchar_t *var_name, int scope, io_streams_t &streams
             if (i >= 50 && i < vals.size() - 50) continue;
         }
         const wcstring value = vals[i];
-        const wcstring escaped_val = escape_string(value, ESCAPE_NO_QUOTED, STRING_STYLE_SCRIPT);
+        const wcstring escaped_val = escape_string(value, ESCAPE_NO_PRINTABLES | ESCAPE_NO_QUOTED);
         streams.out.append_format(_(L"$%ls[%d]: |%ls|\n"), var_name, i + 1, escaped_val.c_str());
     }
 }
@@ -556,14 +568,24 @@ static int builtin_set_show(const wchar_t *cmd, const set_cmd_opts_t &opts, int
                             const wchar_t **argv, parser_t &parser, io_streams_t &streams) {
     UNUSED(opts);
     const auto &vars = parser.vars();
+    auto inheriteds = env_get_inherited();
     if (argc == 0) {  // show all vars
-        wcstring_list_t names = parser.vars().get_names(ENV_USER);
+        wcstring_list_t names = vars.get_names(ENV_USER);
         sort(names.begin(), names.end());
         for (const auto &name : names) {
             if (name == L"history") continue;
             show_scope(name.c_str(), ENV_LOCAL, streams, vars);
             show_scope(name.c_str(), ENV_GLOBAL, streams, vars);
             show_scope(name.c_str(), ENV_UNIVERSAL, streams, vars);
+
+            // Show the originally imported value as a debugging aid.
+            auto inherited = inheriteds.find(name);
+            if (inherited != inheriteds.end()) {
+                const wcstring escaped_val =
+                    escape_string(inherited->second, ESCAPE_NO_PRINTABLES | ESCAPE_NO_QUOTED);
+                streams.out.append_format(_(L"$%ls: originally inherited as |%ls|\n"), name.c_str(),
+                                          escaped_val.c_str());
+            }
         }
     } else {
         for (int i = 0; i < argc; i++) {
@@ -585,6 +607,13 @@ static int builtin_set_show(const wchar_t *cmd, const set_cmd_opts_t &opts, int
             show_scope(arg, ENV_LOCAL, streams, vars);
             show_scope(arg, ENV_GLOBAL, streams, vars);
             show_scope(arg, ENV_UNIVERSAL, streams, vars);
+            auto inherited = inheriteds.find(arg);
+            if (inherited != inheriteds.end()) {
+                const wcstring escaped_val =
+                    escape_string(inherited->second, ESCAPE_NO_PRINTABLES | ESCAPE_NO_QUOTED);
+                streams.out.append_format(_(L"$%ls: originally inherited as |%ls|\n"), arg,
+                                          escaped_val.c_str());
+            }
         }
     }
 
@@ -595,45 +624,48 @@ static int builtin_set_show(const wchar_t *cmd, const set_cmd_opts_t &opts, int
 static int builtin_set_erase(const wchar_t *cmd, set_cmd_opts_t &opts, int argc,
                              const wchar_t **argv, parser_t &parser, io_streams_t &streams) {
     int ret = STATUS_CMD_OK;
-    env_mode_flags_t scope = compute_scope(opts);
-    for (int i = 0; i < argc; i++) {
-        auto split = split_var_and_indexes(argv[i], scope, parser.vars(), streams);
-        if (!split) {
-            builtin_print_error_trailer(parser, streams.err, cmd);
-            return STATUS_CMD_ERROR;
-        }
-
-        if (!valid_var_name(split->varname)) {
-            streams.err.append_format(BUILTIN_ERR_VARNAME, cmd, split->varname.c_str());
-            builtin_print_error_trailer(parser, streams.err, cmd);
-            return STATUS_INVALID_ARGS;
-        }
-
-        int retval = STATUS_CMD_OK;
-        if (split->indexes.empty()) {  // unset the var
-            retval = parser.vars().remove(split->varname, scope);
-            // When a non-existent-variable is unset, return ENV_NOT_FOUND as $status
-            // but do not emit any errors at the console as a compromise between user
-            // friendliness and correctness.
-            if (retval != ENV_NOT_FOUND) {
-                handle_env_return(retval, cmd, split->varname, streams);
+    env_mode_flags_t scopes = compute_scope(opts);
+    // `set -e` is allowed to be called with multiple scopes.
+    for (int bit = 0; 1 << bit <= ENV_USER; ++bit) {
+        int scope = scopes & 1 << bit;
+        if (scope == 0 || (scope == ENV_USER && scopes != ENV_USER)) continue;
+        for (int i = 0; i < argc; i++) {
+            auto split = split_var_and_indexes(argv[i], scope, parser.vars(), streams);
+            if (!split) {
+                builtin_print_error_trailer(parser, streams.err, cmd);
+                return STATUS_CMD_ERROR;
             }
-            if (retval == ENV_OK) {
-                event_fire(parser, event_t::variable_erase(split->varname));
-            }
-        } else {  // remove just the specified indexes of the var
-            if (!split->var) return STATUS_CMD_ERROR;
-            wcstring_list_t result = erased_at_indexes(split->var->as_list(), split->indexes);
-            retval = env_set_reporting_errors(cmd, split->varname, scope, std::move(result),
-                                              streams, parser);
-        }
 
-        // Set $status to the last error value.
-        // This is cheesy, but I don't expect this to be checked often.
-        if (retval != STATUS_CMD_OK) {
-            ret = retval;
-        } else {
-            warn_if_uvar_shadows_global(cmd, opts, split->varname, streams, parser);
+            if (!valid_var_name(split->varname)) {
+                streams.err.append_format(BUILTIN_ERR_VARNAME, cmd, split->varname.c_str());
+                builtin_print_error_trailer(parser, streams.err, cmd);
+                return STATUS_INVALID_ARGS;
+            }
+
+            int retval = STATUS_CMD_OK;
+            if (split->indexes.empty()) {  // unset the var
+                retval = parser.vars().remove(split->varname, scope);
+                // When a non-existent-variable is unset, return ENV_NOT_FOUND as $status
+                // but do not emit any errors at the console as a compromise between user
+                // friendliness and correctness.
+                if (retval != ENV_NOT_FOUND) {
+                    handle_env_return(retval, cmd, split->varname, streams);
+                }
+                if (retval == ENV_OK) {
+                    event_fire(parser, event_t::variable_erase(split->varname));
+                }
+            } else {  // remove just the specified indexes of the var
+                if (!split->var) return STATUS_CMD_ERROR;
+                wcstring_list_t result = erased_at_indexes(split->var->as_list(), split->indexes);
+                retval = env_set_reporting_errors(cmd, split->varname, scope, std::move(result),
+                                                  streams, parser);
+            }
+
+            // Set $status to the last error value.
+            // This is cheesy, but I don't expect this to be checked often.
+            if (retval != STATUS_CMD_OK) {
+                ret = retval;
+            }
         }
     }
     return ret;
@@ -720,6 +752,12 @@ static int builtin_set_set(const wchar_t *cmd, set_cmd_opts_t &opts, int argc, c
     // Is the variable valid?
     if (!valid_var_name(split->varname)) {
         streams.err.append_format(BUILTIN_ERR_VARNAME, cmd, split->varname.c_str());
+        auto pos = split->varname.find(L'=');
+        if (pos != wcstring::npos) {
+            streams.err.append_format(L"%ls: Did you mean `set %ls %ls`?", cmd,
+                                      escape_string(split->varname.substr(0, pos)).c_str(),
+                                      escape_string(split->varname.substr(pos + 1)).c_str());
+        }
         builtin_print_error_trailer(parser, streams.err, cmd);
         return STATUS_INVALID_ARGS;
     }
@@ -759,11 +797,13 @@ static int builtin_set_set(const wchar_t *cmd, set_cmd_opts_t &opts, int argc, c
         new_values = new_var_values_by_index(*split, argc, argv);
     }
 
-    warn_if_uvar_shadows_global(cmd, opts, split->varname, streams, parser);
     // Set the value back in the variable stack and fire any events.
     int retval = env_set_reporting_errors(cmd, split->varname, scope, std::move(new_values),
                                           streams, parser);
 
+    if (retval == ENV_OK) {
+        warn_if_uvar_shadows_global(cmd, opts, split->varname, streams, parser);
+    }
     return retval;
 }
 
@@ -784,7 +824,7 @@ maybe_t<int> builtin_set(parser_t &parser, io_streams_t &streams, const wchar_t
         return STATUS_CMD_OK;
     }
 
-    retval = validate_cmd_opts(cmd, opts, argc, parser, streams);
+    retval = validate_cmd_opts(cmd, opts, argc, argv, parser, streams);
     if (retval != STATUS_CMD_OK) return retval;
 
     if (opts.query) {
