@@ -3,16 +3,24 @@
 
 #include "wait.h"
 
-#include <sys/wait.h>
-
 #include <algorithm>
+#include <cerrno>
+#include <csignal>
+#include <deque>
+#include <list>
+#include <memory>
+#include <string>
+#include <utility>
 #include <vector>
 
 #include "../builtin.h"
 #include "../common.h"
+#include "../io.h"
+#include "../maybe.h"
 #include "../parser.h"
 #include "../proc.h"
 #include "../signal.h"
+#include "../topic_monitor.h"
 #include "../wait_handle.h"
 #include "../wgetopt.h"
 #include "../wutil.h"
@@ -129,7 +137,7 @@ maybe_t<int> builtin_wait(parser_t &parser, io_streams_t &streams, const wchar_t
 
     static const wchar_t *const short_options = L":nh";
     static const struct woption long_options[] = {
-        {L"any", no_argument, nullptr, 'n'}, {L"help", no_argument, nullptr, 'h'}, {}};
+        {L"any", no_argument, 'n'}, {L"help", no_argument, 'h'}, {}};
 
     int opt;
     wgetopter_t w;
