@@ -1,12 +1,9 @@
 // String expansion functions. These functions perform several kinds of parameter expansion.
-// IWYU pragma: no_include <cstddef>
-#include "config.h"
+#include "config.h"  // IWYU pragma: keep
 
 #include <errno.h>
 #include <pwd.h>
 #include <stdarg.h>
-#include <stddef.h>
-#include <stdlib.h>
 #include <unistd.h>
 #include <wctype.h>
 
@@ -16,20 +13,11 @@
 #ifdef SunOS
 #include <procfs.h>
 #endif
-#ifdef __APPLE__
-#include <sys/time.h>  // Required to build with old SDK versions
-// proc.h needs to be included *after* time.h, this comment stops clang-format from reordering.
-#include <sys/proc.h>
-#else
-#include <dirent.h>
-#include <sys/stat.h>
-#endif
 
 #include <algorithm>
-#include <functional>
 #include <map>
-#include <memory>  // IWYU pragma: keep
-#include <type_traits>
+#include <memory>
+#include <unordered_map>
 #include <utility>
 #include <vector>
 
@@ -40,13 +28,11 @@
 #include "expand.h"
 #include "fallback.h"  // IWYU pragma: keep
 #include "history.h"
-#include "iothread.h"
+#include "operation_context.h"
 #include "parse_constants.h"
 #include "parse_util.h"
 #include "parser.h"
 #include "path.h"
-#include "proc.h"
-#include "reader.h"
 #include "util.h"
 #include "wcstringutil.h"
 #include "wildcard.h"
@@ -98,13 +84,13 @@ static void append_syntax_error(parse_error_list_t *errors, size_t source_start,
 /// Append a cmdsub error to the given error list. But only do so if the error hasn't already been
 /// recorded. This is needed because command substitution is a recursive process and some errors
 /// could consequently be recorded more than once.
-static void append_cmdsub_error(parse_error_list_t *errors, size_t source_start, const wchar_t *fmt,
-                                ...) {
+static void append_cmdsub_error(parse_error_list_t *errors, size_t source_start, size_t source_end,
+                                const wchar_t *fmt, ...) {
     if (!errors) return;
 
     parse_error_t error;
     error.source_start = source_start;
-    error.source_length = 0;
+    error.source_length = source_end - source_start + 1;
     error.code = parse_error_cmdsubst;
 
     va_list va;
@@ -154,7 +140,7 @@ wcstring expand_escape_variable(const env_var_t &var) {
             buff.append(el);
             buff.append(L"'");
         } else {
-            buff.append(escape_string(el, 1));
+            buff.append(escape_string(el));
         }
     }
 
@@ -169,7 +155,7 @@ wcstring expand_escape_string(const wcstring &el) {
         buff.append(el);
         buff.append(L"'");
     } else {
-        buff.append(escape_string(el, 1));
+        buff.append(escape_string(el));
     }
     return buff;
 }
@@ -186,7 +172,7 @@ enum class parse_slice_error_t {
 static size_t parse_slice(const wchar_t *const in, wchar_t **const end_ptr, std::vector<long> &idx,
                           size_t array_size, parse_slice_error_t *const error) {
     const long size = static_cast<long>(array_size);
-    size_t pos = 1;  // skip past the opening square brace
+    size_t pos = 1;  // skip past the opening square bracket
 
     *error = parse_slice_error_t::none;
 
@@ -290,7 +276,7 @@ static size_t parse_slice(const wchar_t *const in, wchar_t **const end_ptr, std:
 ///
 /// This function operates on strings backwards, starting at last_idx.
 ///
-/// Note: last_idx is considered to be where it previously finished procesisng. This means it
+/// Note: last_idx is considered to be where it previously finished processing. This means it
 /// actually starts operating on last_idx-1. As such, to process a string fully, pass string.size()
 /// as last_idx instead of string.size()-1.
 ///
@@ -571,6 +557,7 @@ static expand_result_t expand_braces(wcstring &&instr, expand_flags_t flags,
     }
 
     if (brace_begin == nullptr) {
+        // No more brace expansions left; we can return the value as-is.
         if (!out->add(std::move(instr))) {
             return expand_result_t::error;
         }
@@ -593,6 +580,11 @@ static expand_result_t expand_braces(wcstring &&instr, expand_flags_t flags,
                 }
             }
 
+            // `whole_item` is a whitespace- and brace-stripped member of a single pass of brace
+            // expansion, e.g. in `{ alpha , b,{c, d }}`, `alpha`, `b`, and `c, d` will, in the
+            // first round of expansion, each in turn be a `whole_item` (with recursive commas
+            // replaced by special placeholders).
+            // We recursively call `expand_braces` with each item until it's been fully expanded.
             wcstring whole_item;
             whole_item.reserve(tot_len + item_len + 2);
             whole_item.append(in, length_preceding_braces);
@@ -626,8 +618,9 @@ static expand_result_t expand_cmdsubst(wcstring input, const operation_context_t
     wcstring subcmd;
 
     bool is_quoted = false;
+    bool has_dollar = false;
     switch (parse_util_locate_cmdsubst_range(input, &cursor, &subcmd, &paren_begin, &paren_end,
-                                             false, &is_quoted)) {
+                                             false, &is_quoted, &has_dollar)) {
         case -1: {
             append_syntax_error(errors, SOURCE_LOCATION_UNKNOWN, L"Mismatched parenthesis");
             return expand_result_t::make_error(STATUS_EXPAND_ERROR);
@@ -646,8 +639,6 @@ static expand_result_t expand_cmdsubst(wcstring input, const operation_context_t
         }
     }
 
-    bool have_dollar = paren_begin > 0 && input.at(paren_begin - 1) == L'$';
-
     wcstring_list_t sub_res;
     int subshell_status = exec_subshell_for_expand(subcmd, *ctx.parser, ctx.job_group, sub_res);
     if (subshell_status != 0) {
@@ -657,8 +648,13 @@ static expand_result_t expand_cmdsubst(wcstring input, const operation_context_t
             case STATUS_READ_TOO_MUCH:
                 err = L"Too much data emitted by command substitution so it was discarded";
                 break;
+            // TODO: STATUS_CMD_ERROR is overused and too generic. We shouldn't have to test things
+            // to figure out what error to show after we've already been given an error code.
             case STATUS_CMD_ERROR:
                 err = L"Too many active file descriptors";
+                if (ctx.parser->is_eval_depth_exceeded()) {
+                    err = L"Unable to evaluate string substitution";
+                }
                 break;
             case STATUS_CMD_UNKNOWN:
                 err = L"Unknown command";
@@ -669,11 +665,28 @@ static expand_result_t expand_cmdsubst(wcstring input, const operation_context_t
             case STATUS_NOT_EXECUTABLE:
                 err = L"Command not executable";
                 break;
+            case STATUS_INVALID_ARGS:
+                // TODO: Also overused
+                // This is sent for:
+                // invalid redirections or pipes (like `<&foo`),
+                // invalid variables (invalid name or read-only) for for-loops,
+                // switch $foo if $foo expands to more than one argument
+                // time in a background job.
+                err = L"Invalid arguments";
+                break;
+            case STATUS_EXPAND_ERROR:
+                // Sent in `for $foo in ...` if $foo expands to more than one word
+                err = L"Expansion error";
+                break;
+            case STATUS_UNMATCHED_WILDCARD:
+                // Sent in `for $foo in ...` if $foo expands to more than one word
+                err = L"Unmatched wildcard";
+                break;
             default:
                 err = L"Unknown error while evaluating command substitution";
                 break;
         }
-        append_cmdsub_error(errors, paren_begin, _(err));
+        append_cmdsub_error(errors, paren_begin, paren_end, _(err));
         return expand_result_t::make_error(subshell_status);
     }
 
@@ -757,8 +770,10 @@ static expand_result_t expand_cmdsubst(wcstring input, const operation_context_t
             wcstring whole_item;
             whole_item.reserve(paren_begin + 1 + sub_res_joined.size() + 1 +
                                tail_item.completion.size());
-            whole_item.append(input, 0, paren_begin - have_dollar);
+            whole_item.append(input, 0, paren_begin - has_dollar);
+            whole_item.push_back(INTERNAL_SEPARATOR);
             whole_item.append(sub_res_joined);
+            whole_item.push_back(INTERNAL_SEPARATOR);
             whole_item.append(tail_item.completion.substr(const_strlen(L"\"")));
             if (!out->add(std::move(whole_item))) {
                 return append_overflow_error(errors);
@@ -769,12 +784,12 @@ static expand_result_t expand_cmdsubst(wcstring input, const operation_context_t
     }
 
     for (const wcstring &sub_item : sub_res) {
-        wcstring sub_item2 = escape_string(sub_item, ESCAPE_ALL);
+        wcstring sub_item2 = escape_string(sub_item);
         for (const completion_t &tail_item : tail_expand) {
             wcstring whole_item;
             whole_item.reserve(paren_begin + 1 + sub_item2.size() + 1 +
                                tail_item.completion.size());
-            whole_item.append(input, 0, paren_begin - have_dollar);
+            whole_item.append(input, 0, paren_begin - has_dollar);
             whole_item.push_back(INTERNAL_SEPARATOR);
             whole_item.append(sub_item2);
             whole_item.push_back(INTERNAL_SEPARATOR);
@@ -858,43 +873,6 @@ void expand_tilde(wcstring &input, const environment_t &vars) {
     }
 }
 
-static void unexpand_tildes(const wcstring &input, const environment_t &vars,
-                            completion_list_t *completions) {
-    // If input begins with tilde, then try to replace the corresponding string in each completion
-    // with the tilde. If it does not, there's nothing to do.
-    if (input.empty() || input.at(0) != L'~') return;
-
-    // We only operate on completions that replace their contents. If we don't have any, we're done.
-    // In particular, empty vectors are common.
-    bool has_candidate_completion = false;
-    for (const auto &completion : *completions) {
-        if (completion.flags & COMPLETE_REPLACES_TOKEN) {
-            has_candidate_completion = true;
-            break;
-        }
-    }
-    if (!has_candidate_completion) return;
-
-    size_t tail_idx;
-    wcstring username_with_tilde = L"~";
-    username_with_tilde.append(get_home_directory_name(input, &tail_idx));
-
-    // Expand username_with_tilde.
-    wcstring home = username_with_tilde;
-    expand_tilde(home, vars);
-
-    // Now for each completion that starts with home, replace it with the username_with_tilde.
-    for (auto &comp : *completions) {
-        if ((comp.flags & COMPLETE_REPLACES_TOKEN) &&
-            string_prefixes_string(home, comp.completion)) {
-            comp.completion.replace(0, home.size(), username_with_tilde);
-
-            // And mark that our tilde is literal, so it doesn't try to escape it.
-            comp.flags |= COMPLETE_DONT_ESCAPE_TILDES;
-        }
-    }
-}
-
 // If the given path contains the user's home directory, replace that with a tilde. We don't try to
 // be smart about case insensitivity, etc.
 wcstring replace_home_directory_with_tilde(const wcstring &str, const environment_t &vars) {
@@ -903,6 +881,9 @@ wcstring replace_home_directory_with_tilde(const wcstring &str, const environmen
     if (string_prefixes_string(L"/", result)) {
         wcstring home_directory = L"~";
         expand_tilde(home_directory, vars);
+        // If we can't get a home directory, don't replace anything.
+        // This is the case e.g. with --no-execute.
+        if (home_directory.empty()) return result;
         if (!string_suffixes_string(L"/", home_directory)) {
             home_directory.push_back(L'/');
         }
@@ -971,6 +952,10 @@ class expander_t {
     expander_t(const operation_context_t &ctx, expand_flags_t flags, parse_error_list_t *errors)
         : ctx(ctx), flags(flags), errors(errors) {}
 
+    // Given an original input string, if it starts with a tilde, "unexpand" the expanded home
+    // directory.  Note this may be just a tilde or a user name like ~foo/.
+    void unexpand_tildes(const wcstring &input, completion_list_t *completions) const;
+
    public:
     static expand_result_t expand_string(wcstring input, completion_receiver_t *out_completions,
                                          expand_flags_t flags, const operation_context_t &ctx,
@@ -987,8 +972,8 @@ expand_result_t expander_t::stage_cmdsubst(wcstring input, completion_receiver_t
                 }
                 return expand_result_t::ok;
             case 1:
-                append_cmdsub_error(errors, start, L"Command substitutions not allowed");
-                /* intentionally falls through */
+                append_cmdsub_error(errors, start, end, L"command substitutions not allowed here");
+                __fallthrough__
             case -1:
             default:
                 return expand_result_t::make_error(STATUS_EXPAND_ERROR);
@@ -1008,7 +993,7 @@ expand_result_t expander_t::stage_variables(wcstring input, completion_receiver_
 
     if (flags & expand_flag::skip_variables) {
         for (auto &i : next) {
-            if (i == VARIABLE_EXPAND) {
+            if (i == VARIABLE_EXPAND || i == VARIABLE_EXPAND_SINGLE) {
                 i = L'$';
             }
         }
@@ -1137,6 +1122,50 @@ expand_result_t expander_t::stage_wildcards(wcstring path_to_expand, completion_
     return result;
 }
 
+void expander_t::unexpand_tildes(const wcstring &input, completion_list_t *completions) const {
+    // If input begins with tilde, then try to replace the corresponding string in each completion
+    // with the tilde. If it does not, there's nothing to do.
+    if (input.empty() || input.at(0) != L'~') return;
+
+    // This is a subtle kludge. We need to decide whether to unexpand tildes for all
+    // completions, or only those which replace their tokens. The problem is that we're sloppy
+    // about setting the COMPLETE_REPLACES_TOKEN flag, except when we're completing in the
+    // wildcard stage, because no other clients of string expansion care. Example:
+    //   HOME=/foo
+    //   mkdir ~/foo # makes /foo/foo
+    //   cd ~/<tab>
+    // Here we are likely to get a completion 'foo' which may match $HOME, but it extends its token
+    // instead of replacing it, so we don't modify it (it will just be appended to the original ~/).
+    //
+    // However if we are not completing, just expanding, then expansion just produces the full paths
+    // so we should unconditionally unexpand tildes.
+    bool only_replacers = bool(flags & expand_flag::for_completions);
+
+    // Helper to decide whether to process a completion.
+    auto should_process = [=](const completion_t &c) {
+        return only_replacers ? c.replaces_token() : true;
+    };
+
+    // Early out if none qualify.
+    if (std::none_of(completions->begin(), completions->end(), should_process)) return;
+
+    // Get the username_with_tilde (like ~bert) and expand it into a home directory.
+    size_t tail_idx;
+    wcstring username_with_tilde = L"~" + get_home_directory_name(input, &tail_idx);
+    wcstring home = username_with_tilde;
+    expand_tilde(home, ctx.vars);
+
+    // Now for each completion that starts with home, replace it with the username_with_tilde.
+    for (auto &comp : *completions) {
+        if (should_process(comp) && string_prefixes_string(home, comp.completion)) {
+            comp.completion.replace(0, home.size(), username_with_tilde);
+
+            // And mark that our tilde is literal, so it doesn't try to escape it.
+            comp.flags |= COMPLETE_DONT_ESCAPE_TILDES;
+        }
+    }
+}
+
 expand_result_t expander_t::expand_string(wcstring input, completion_receiver_t *out_completions,
                                           expand_flags_t flags, const operation_context_t &ctx,
                                           parse_error_list_t *errors) {
@@ -1196,8 +1225,10 @@ expand_result_t expander_t::expand_string(wcstring input, completion_receiver_t
     }
 
     if (total_result == expand_result_t::ok) {
-        // Hack to un-expand tildes (see #647).
-        unexpand_tildes(input, ctx.vars, &completions);
+        // Unexpand tildes if we want to preserve them (see #647).
+        if (flags.get(expand_flag::preserve_home_tildes)) {
+            expand.unexpand_tildes(input, &completions);
+        }
         if (!out_completions->add_list(std::move(completions))) {
             total_result = append_overflow_error(errors);
         }
@@ -1315,31 +1346,3 @@ bool fish_xdm_login_hack_hack_hack_hack(std::vector<std::string> *cmds, int argc
     }
     return result;
 }
-
-maybe_t<wcstring> expand_abbreviation(const wcstring &src, const environment_t &vars) {
-    if (src.empty()) return none();
-
-    wcstring esc_src = escape_string(src, 0, STRING_STYLE_VAR);
-    if (esc_src.empty()) {
-        return none();
-    }
-    wcstring var_name = L"_fish_abbr_" + esc_src;
-    if (auto var_value = vars.get(var_name)) {
-        return var_value->as_string();
-    }
-    return none();
-}
-
-std::map<wcstring, wcstring> get_abbreviations(const environment_t &vars) {
-    const wcstring prefix = L"_fish_abbr_";
-    auto names = vars.get_names(0);
-    std::map<wcstring, wcstring> result;
-    for (const wcstring &name : names) {
-        if (string_prefixes_string(prefix, name)) {
-            wcstring key;
-            unescape_string(name.substr(prefix.size()), &key, UNESCAPE_DEFAULT, STRING_STYLE_VAR);
-            result[key] = vars.get(name)->as_string();
-        }
-    }
-    return result;
-}
