@@ -25,6 +25,17 @@ struct source_range_t {
         assert(start + length >= start && "Overflow");
         return start + length;
     }
+
+    bool operator==(const source_range_t &rhs) const {
+        return start == rhs.start && length == rhs.length;
+    }
+
+    bool operator!=(const source_range_t &rhs) const { return !(*this == rhs); }
+
+    // \return true if a location is in this range, including one-past-the-end.
+    bool contains_inclusive(source_offset_t loc) const {
+        return start <= loc && loc - start <= length;
+    }
 };
 
 // IMPORTANT: If the following enum table is modified you must also update token_enum_map below.
@@ -205,14 +216,21 @@ enum class pipeline_position_t : uint8_t {
 /// Maximum number of function calls.
 #define FISH_MAX_STACK_DEPTH 128
 
+/// Maximum number of nested string substitutions (in lieu of evals)
+/// Reduced under TSAN: our CI test creates 500 jobs and this is very slow with TSAN.
+#if FISH_TSAN_WORKAROUNDS
+#define FISH_MAX_EVAL_DEPTH 250
+#else
+#define FISH_MAX_EVAL_DEPTH 500
+#endif
+
 /// Error message on a function that calls itself immediately.
 #define INFINITE_FUNC_RECURSION_ERR_MSG \
     _(L"The function '%ls' calls itself immediately, which would result in an infinite loop.")
 
 /// Error message on reaching maximum call stack depth.
-#define CALL_STACK_LIMIT_EXCEEDED_ERR_MSG                                                     \
-    _(L"The function call stack limit has been exceeded. Do you have an accidental infinite " \
-      L"loop?")
+#define CALL_STACK_LIMIT_EXCEEDED_ERR_MSG \
+    _(L"The call stack limit has been exceeded. Do you have an accidental infinite loop?")
 
 /// Error message when encountering an unknown builtin name.
 #define UNKNOWN_BUILTIN_ERR_MSG _(L"Unknown builtin '%ls'")
