@@ -38,9 +38,9 @@
 // Ave, Cambridge, MA 02139, USA.
 #include "config.h"  // IWYU pragma: keep
 
+#include <stddef.h>
 #include <stdio.h>
 
-#include <cstring>
 #include <cwchar>
 
 // This version of `getopt' appears to the caller like standard Unix `getopt' but it behaves
@@ -65,11 +65,11 @@
 //
 // `first_nonopt' and `last_nonopt' are relocated so that they describe the new indices of the
 // non-options in ARGV after they are moved.
-void wgetopter_t::exchange(wchar_t **argv) {
+void wgetopter_t::exchange(string_array_t argv) {
     int bottom = first_nonopt;
     int middle = last_nonopt;
     int top = woptind;
-    wchar_t *tem;
+    const wchar_t *tem{};
 
     // Exchange the shorter segment with the far end of the longer segment. That puts the shorter
     // segment into the right place. It leaves the longer segment in the right place overall, but it
@@ -139,7 +139,7 @@ void wgetopter_t::_wgetopt_initialize(const wchar_t *optstring) {
 
 // Advance to the next ARGV-element.
 int wgetopter_t::_advance_to_next_argv(  //!OCLINT(high cyclomatic complexity)
-    int argc, wchar_t **argv, const struct woption *longopts) {
+    int argc, string_array_t argv, const struct woption *longopts) {
     if (ordering == PERMUTE) {
         // If we have just processed some options following some non-options, exchange them so
         // that the options come first.
@@ -196,7 +196,7 @@ int wgetopter_t::_advance_to_next_argv(  //!OCLINT(high cyclomatic complexity)
 }
 
 // Check for a matching short opt.
-int wgetopter_t::_handle_short_opt(int argc, wchar_t **argv) {
+int wgetopter_t::_handle_short_opt(int argc, string_array_t argv) {
     // Look at and handle the next short option-character.
     wchar_t c = *nextchar++;
     const wchar_t *temp = std::wcschr(shortopts, c);
@@ -205,10 +205,6 @@ int wgetopter_t::_handle_short_opt(int argc, wchar_t **argv) {
     if (*nextchar == '\0') ++woptind;
 
     if (temp == nullptr || c == ':') {
-        if (wopterr) {
-            std::fwprintf(stderr, _(L"%ls: Invalid option -- %lc\n"), argv[0],
-                          static_cast<wint_t>(c));
-        }
         woptopt = c;
 
         if (*nextchar != '\0') woptind++;
@@ -236,11 +232,6 @@ int wgetopter_t::_handle_short_opt(int argc, wchar_t **argv) {
             // the next element now.
             woptind++;
         } else if (woptind == argc) {
-            if (wopterr) {
-                // 1003.2 specifies the format of this message.
-                std::fwprintf(stderr, _(L"%ls: Option requires an argument -- %lc\n"), argv[0],
-                              static_cast<wint_t>(c));
-            }
             woptopt = c;
             c = missing_arg_return_colon ? ':' : '?';
         } else {
@@ -254,35 +245,22 @@ int wgetopter_t::_handle_short_opt(int argc, wchar_t **argv) {
     return c;
 }
 
-void wgetopter_t::_update_long_opt(int argc, wchar_t **argv, const struct woption *pfound,
-                                   wchar_t *nameend, int *longind, int option_index, int *retval) {
+void wgetopter_t::_update_long_opt(int argc, string_array_t argv, const struct woption *pfound,
+                                   size_t nameend, int *longind, int option_index, int *retval) {
     woptind++;
-    if (*nameend) {
-        // Don't test has_arg with >, because some C compilers don't allow it to be used on
-        // enums.
-        if (pfound->has_arg)
-            woptarg = nameend + 1;
+    assert(nextchar[nameend] == '\0' || nextchar[nameend] == '=');
+    if (nextchar[nameend] == '=') {
+        if (pfound->has_arg != no_argument)
+            woptarg = &nextchar[nameend + 1];
         else {
-            if (wopterr) {
-                if (argv[woptind - 1][1] == '-')  // --option
-                    std::fwprintf(stderr, _(L"%ls: Option '--%ls' doesn't allow an argument\n"),
-                                  argv[0], pfound->name);
-                else
-                    // +option or -option
-                    std::fwprintf(stderr, _(L"%ls: Option '%lc%ls' doesn't allow an argument\n"),
-                                  argv[0], argv[woptind - 1][0], pfound->name);
-            }
             nextchar += std::wcslen(nextchar);
             *retval = '?';
             return;
         }
-    } else if (pfound->has_arg == 1) {
+    } else if (pfound->has_arg == required_argument) {
         if (woptind < argc)
             woptarg = argv[woptind++];
         else {
-            if (wopterr)
-                std::fwprintf(stderr, _(L"%ls: Option '%ls' requires an argument\n"), argv[0],
-                              argv[woptind - 1]);
             nextchar += std::wcslen(nextchar);
             *retval = missing_arg_return_colon ? ':' : '?';
             return;
@@ -291,26 +269,20 @@ void wgetopter_t::_update_long_opt(int argc, wchar_t **argv, const struct woptio
 
     nextchar += std::wcslen(nextchar);
     if (longind != nullptr) *longind = option_index;
-    if (pfound->flag) {
-        *(pfound->flag) = pfound->val;
-        *retval = 0;
-    } else {
-        *retval = pfound->val;
-    }
+    *retval = pfound->val;
 }
 
 // Find a matching long opt.
 const struct woption *wgetopter_t::_find_matching_long_opt(const struct woption *longopts,
-                                                           wchar_t *nameend, int *exact, int *ambig,
+                                                           size_t nameend, int *exact, int *ambig,
                                                            int *indfound) const {
     const struct woption *pfound = nullptr;
     int option_index = 0;
 
     // Test all long options for either exact match or abbreviated matches.
     for (const struct woption *p = longopts; p->name; p++, option_index++) {
-        if (!std::wcsncmp(p->name, nextchar, nameend - nextchar)) {
-            if (static_cast<unsigned int>(nameend - nextchar) ==
-                static_cast<unsigned int>(wcslen(p->name))) {
+        if (!std::wcsncmp(p->name, nextchar, nameend)) {
+            if (nameend == wcslen(p->name)) {
                 // Exact match found.
                 pfound = p;
                 *indfound = option_index;
@@ -330,23 +302,21 @@ const struct woption *wgetopter_t::_find_matching_long_opt(const struct woption
 }
 
 // Check for a matching long opt.
-bool wgetopter_t::_handle_long_opt(int argc, wchar_t **argv, const struct woption *longopts,
+bool wgetopter_t::_handle_long_opt(int argc, string_array_t argv, const struct woption *longopts,
                                    int *longind, int long_only, int *retval) {
     int exact = 0;
     int ambig = 0;
     int indfound = 0;
 
-    wchar_t *nameend;
-    for (nameend = nextchar; *nameend && *nameend != '='; nameend++)
-        ;  //!OCLINT(empty body)
+    size_t nameend = 0;
+    while (nextchar[nameend] && nextchar[nameend] != '=') {
+        nameend++;
+    }
 
     const struct woption *pfound =
         _find_matching_long_opt(longopts, nameend, &exact, &ambig, &indfound);
 
     if (ambig && !exact) {
-        if (wopterr) {
-            std::fwprintf(stderr, _(L"%ls: Option '%ls' is ambiguous\n"), argv[0], argv[woptind]);
-        }
         nextchar += std::wcslen(nextchar);
         woptind++;
         *retval = '?';
@@ -362,14 +332,6 @@ bool wgetopter_t::_handle_long_opt(int argc, wchar_t **argv, const struct woptio
     // with '--' or is not a valid short option, then it's an error. Otherwise interpret it as a
     // short option.
     if (!long_only || argv[woptind][1] == '-' || std::wcschr(shortopts, *nextchar) == nullptr) {
-        if (wopterr) {
-            if (argv[woptind][1] == '-')  // --option
-                std::fwprintf(stderr, _(L"%ls: Unrecognized option '--%ls'\n"), argv[0], nextchar);
-            else
-                // +option or -option
-                std::fwprintf(stderr, _(L"%ls: Unrecognized option '%lc%ls'\n"), argv[0],
-                              argv[woptind][0], nextchar);
-        }
         nextchar = const_cast<wchar_t *>(L"");
         woptind++;
         *retval = '?';
@@ -395,8 +357,7 @@ bool wgetopter_t::_handle_long_opt(int argc, wchar_t **argv, const struct woptio
 // that those that are not options now come last.)
 //
 // OPTSTRING is a string containing the legitimate option characters. If an option character is seen
-// that is not listed in OPTSTRING, return '?' after printing an error message.  If you set
-// `wopterr' to zero, the error message is suppressed but we still return '?'.
+// that is not listed in OPTSTRING, return '?'.
 //
 // If a char in OPTSTRING is followed by a colon, that means it wants an arg, so the following text
 // in the same ARGV-element, or the text of the following ARGV-element, is returned in `optarg'.
@@ -419,7 +380,7 @@ bool wgetopter_t::_handle_long_opt(int argc, wchar_t **argv, const struct woptio
 // long-named option has been found by the most recent call.
 //
 // If LONG_ONLY is nonzero, '-' as well as '--' can introduce long-named options.
-int wgetopter_t::_wgetopt_internal(int argc, wchar_t **argv, const wchar_t *optstring,
+int wgetopter_t::_wgetopt_internal(int argc, string_array_t argv, const wchar_t *optstring,
                                    const struct woption *longopts, int *longind, int long_only) {
     if (!initialized) _wgetopt_initialize(optstring);
     woptarg = nullptr;
@@ -442,17 +403,32 @@ int wgetopter_t::_wgetopt_internal(int argc, wchar_t **argv, const wchar_t *opts
     // "u".
     //
     // This distinction seems to be the most useful approach.
-    if (longopts != nullptr &&
-        (argv[woptind][1] == '-' ||
-         (long_only && (argv[woptind][2] || !std::wcschr(shortopts, argv[woptind][1]))))) {
-        int retval;
-        if (_handle_long_opt(argc, argv, longopts, longind, long_only, &retval)) return retval;
+    if (longopts && woptind < argc) {
+        const wchar_t *arg = argv[woptind];
+        assert(arg && "Null arg");
+        bool try_long = false;
+        if (arg[0] == '-' && arg[1] == '-') {
+            // Like --foo
+            try_long = true;
+        } else if (long_only && wcslen(arg) >= 3) {
+            // Like -fu
+            try_long = true;
+        } else if (!std::wcschr(shortopts, arg[1])) {
+            // Like -f, but f is not a short arg.
+            try_long = true;
+        }
+        if (try_long) {
+            int retval = 0;
+            if (_handle_long_opt(argc, argv, longopts, longind, long_only, &retval)) {
+                return retval;
+            }
+        }
     }
-
     return _handle_short_opt(argc, argv);
 }
 
-int wgetopter_t::wgetopt_long(int argc, wchar_t **argv, const wchar_t *options,
+int wgetopter_t::wgetopt_long(int argc, string_array_t argv, const wchar_t *options,
                               const struct woption *long_options, int *opt_index) {
+    assert(woptind <= argc && "woptind is out of range");
     return _wgetopt_internal(argc, argv, options, long_options, opt_index, 0);
 }
